/*
 * Copyright (C) 2020 Loongson Corporation
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 * Authors:
 *    Sui Jingfeng <suijingfeng@loongson.cn>
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <errno.h>
#include <unistd.h>
#include <malloc.h>

#include "dumb_bo.h"
#include "driver.h"
#include "loongson_scanout.h"
#include "loongson_pixmap.h"

/* gsgpu */
#include <gsgpu_drm.h>
#include <gsgpu.h>
//
// NOTE: loongson's front bo(scanout buffer) is a dumb,
// not a gbm bo, we want it clear
Bool LS_CreateFrontBO(ScrnInfoPtr pScrn, struct drmmode_rec * const pDrmMode)
{
    int bpp = pDrmMode->kbpp;
    int cpp = (bpp + 7) / 8;
    int width = pScrn->virtualX;
    int height = pScrn->virtualY;
    struct DrmModeBO * const pFrontBO = &pDrmMode->front_bo;

    pFrontBO->dumb = dumb_bo_create(pDrmMode->fd, width, height, bpp);
    if (pFrontBO->dumb == NULL)
    {
        xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
                   "%s: create dumb BO(%dx%d, bpp=%d) failed\n",
                   __func__, width, height, bpp);
        return FALSE;
    }

    pFrontBO->width = width;
    pFrontBO->height = height;

    /* pitch */
    pScrn->displayWidth = pFrontBO->dumb->pitch / cpp;

    xf86DrvMsg(pScrn->scrnIndex, X_INFO,
               "%s: front BO (%dx%d, bpp=%d) created\n",
               __func__, width, height, bpp);

    xf86DrvMsg(pScrn->scrnIndex, X_INFO,
               "CPP (number of bytes per pixel) = %d\n", cpp);
    xf86DrvMsg(pScrn->scrnIndex, X_INFO,
               "Display Width (number of pixels a line) = %d\n",
               pScrn->displayWidth);

    return TRUE;
}

void LS_FreeFrontBO(ScrnInfoPtr pScrn, struct drmmode_rec * const pDrmMode)
{
    int ret;
    drmmode_bo * const pFBO = &pDrmMode->front_bo;

    if (pDrmMode->fb_id)
    {
        drmModeRmFB(pDrmMode->fd, pDrmMode->fb_id);
        xf86DrvMsg(pScrn->scrnIndex, X_INFO,
                "Front BO(fb_id = %d) get removed from FB.\n", pDrmMode->fb_id);
        pDrmMode->fb_id = 0;
    }

    // suijingfeng : check is not necessary ...
    // for loongson-drm, we are sure that the front bo is just a dumb ...
    if (pFBO->dumb)
    {
        ret = dumb_bo_destroy(pDrmMode->fd, pFBO->dumb);
        if (ret == 0)
        {
            pFBO->dumb = NULL;
            xf86DrvMsg(pScrn->scrnIndex, X_INFO,
                "Front BO get freed.\n");
        }
    }
}


// Is a DUMB, not GBM, this is guarenteed.
void *LS_MapFrontBO(ScrnInfoPtr pScrn, struct drmmode_rec * const pDrmMode)
{
    struct DrmModeBO * const pFBO = &pDrmMode->front_bo;
    int ret;

    if (pFBO->dumb->ptr)
    {
        xf86DrvMsg(pScrn->scrnIndex, X_WARNING, "Front BO already Mapped.\n");
        return pFBO->dumb->ptr;
    }

    ret = dumb_bo_map(pDrmMode->fd, pFBO->dumb);
    if (ret)
    {
        xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
                   "%s: Failed map front BO: %d.\n", __func__, ret);
        return NULL;
    }

    xf86DrvMsg(pScrn->scrnIndex, X_INFO, "%s: Front BO Mapped.\n", __func__);

    return pFBO->dumb->ptr;
}


/*
 * Return TRUE if success
 */
Bool loongson_crtc_get_fb_id(xf86CrtcPtr crtc,
                             uint32_t *fb_id,
                             int *x,
                             int *y)
{
    drmmode_crtc_private_ptr drmmode_crtc = crtc->driver_private;
    drmmode_ptr pDrmMode = drmmode_crtc->drmmode;
    PixmapPtr prime_pixmap_ptr = drmmode_crtc->prime_pixmap;
    int ret;

    *fb_id = 0;

    if (prime_pixmap_ptr)
    {
        xf86Msg(X_INFO, "%s: prime_pixmap=%p\n", __func__, prime_pixmap_ptr);

        if (1 /* !drmmode->reverse_prime_offload_mode */)
        {
            msPixmapPrivPtr ppriv = msGetPixmapPriv(pDrmMode, prime_pixmap_ptr);
            *fb_id = ppriv->fb_id;
            *x = 0;
        } else
            *x = drmmode_crtc->prime_pixmap_x;
        *y = 0;
    }
    else if (drmmode_crtc->rotate_fb_id)
    {
        *fb_id = drmmode_crtc->rotate_fb_id;
        *x = *y = 0;
        xf86Msg(X_INFO, "%s: rotate_fb_id=%d\n", __func__,
                drmmode_crtc->rotate_fb_id);
    }
    else
    {
        *fb_id = pDrmMode->fb_id;
        *x = crtc->x;
        *y = crtc->y;

        xf86Msg(X_INFO, "%s: %d: fb_id=%d\n",
                __func__, __LINE__, pDrmMode->fb_id);
    }

    if (*fb_id == 0)
    {
        xf86Msg(X_INFO, "%s: Front bo haven't been scanout, scanout it now\n",
                __func__);

        ret = drmmode_bo_import(pDrmMode,
                                &pDrmMode->front_bo,
                                &pDrmMode->fb_id);
        if (ret < 0)
        {
            xf86Msg(X_ERROR, "failed to add fb %d\n", ret);
            return FALSE;
        }
        *fb_id = pDrmMode->fb_id;
    }

    return TRUE;
}

static struct drmmode_fb *loongson_pixmap_get_fb_ptr(PixmapPtr pPix)
{
    struct exa_pixmap_priv *priv = exaGetPixmapDriverPrivate(pPix);

    if (priv)
        return priv->fb;

    return NULL;
}

static Bool loongson_pixmap_set_fb_ptr(PixmapPtr pPix, struct drmmode_fb *fb)
{
    struct exa_pixmap_priv *priv = exaGetPixmapDriverPrivate(pPix);

    if (!priv)
    {
        return FALSE;
    }

    priv->fb = fb;

    return TRUE;
}

static Bool loongson_pixmap_get_handle(PixmapPtr pPixmap, uint32_t *pHandle)
{
    struct exa_pixmap_priv *priv;

    priv = exaGetPixmapDriverPrivate(pPixmap);
    if (!priv)
    {
        return FALSE;
    }

    if (priv->bo)
    {
        *pHandle = priv->bo->handle;
        return TRUE;
    }

    if (priv->gbo)
    {
        gsgpu_bo_export(priv->gbo, gsgpu_bo_handle_type_kms, pHandle);
        return TRUE;
    }

    return FALSE;
}

static struct drmmode_fb *
loongson_fb_create(ScrnInfoPtr pScrn,
                   int drm_fd,
                   uint32_t width,
                   uint32_t height,
                   uint32_t pitch,
                   uint32_t handle)
{
    struct drmmode_fb *fb = malloc(sizeof(*fb));
    int ret;

    if (!fb)
        return NULL;

    fb->refcnt = 1;

    ret = drmModeAddFB(drm_fd, width, height,
                       pScrn->depth, pScrn->bitsPerPixel,
                       pitch, handle, &fb->id);

    if (ret)
    {
        free(fb);

        xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
                   "%s Failed at %d, return %d\n", __func__, __LINE__, ret);
        return NULL;
    }

    return fb;
}

Bool loongson_pixmap_get_fb_id(PixmapPtr pPixmap, uint32_t *fb_id)
{
    uint32_t handle;
    struct drmmode_fb *fb;
    Bool ret = TRUE;

    fb = loongson_pixmap_get_fb_ptr(pPixmap);
    if (fb)
    {
        *fb_id = fb->id;
        return ret;
    }

    xf86Msg(X_INFO, "%s: don't have fb attach to pixmap(%p), create one\n",
            __func__, pPixmap);

    if (loongson_pixmap_get_handle(pPixmap, &handle))
    {
        ScreenPtr pScreen = pPixmap->drawable.pScreen;
        ScrnInfoPtr pScrn = xf86ScreenToScrn(pScreen);
        loongsonPtr lsp = loongsonPTR(pScrn);

        fb = loongson_fb_create(pScrn,
                                lsp->fd,
                                pPixmap->drawable.width,
                                pPixmap->drawable.height,
                                pPixmap->devKind,
                                handle);

        /* After the fb have been scanout, feed the fb_id to the caller */
        *fb_id = fb->id;

        ret = loongson_pixmap_set_fb_ptr(pPixmap, fb);
    }

    return ret;
}


static void loongson_crtc_scanout_destroy(PixmapPtr *ppScanout)
{
    PixmapPtr pScanout = *ppScanout;
    if (!pScanout)
        return;

    pScanout->drawable.pScreen->DestroyPixmap(pScanout);
    pScanout = NULL;
}


Bool loongson_crtc_create_scanout_pixmap(xf86CrtcPtr pCrtc,
                                         int width,
                                         int height,
                                         PixmapPtr *ppScanoutPix)
{
    ScrnInfoPtr pScrn = pCrtc->scrn;
    ScreenPtr pScreen = pScrn->pScreen;
    PixmapPtr pScanout = *ppScanoutPix;

    if (pScanout)
    {
        if (pScanout->drawable.width == width &&
            pScanout->drawable.height == height)
            return TRUE;

        loongson_crtc_scanout_destroy(ppScanoutPix);
    }

    pScanout = pScreen->CreatePixmap(pScreen,
                                     width,
                                     height,
                                     pScrn->depth,
                                     CREATE_PIXMAP_USAGE_SCANOUT);
    if (!pScanout)
    {
        ErrorF("failed to create CRTC scanout pixmap\n");
        return FALSE;
    }

    *ppScanoutPix = pScanout;

    return TRUE;
}
